unit ecma_stmt;

//`
//2001/04/13 ~
//by Wolfy

interface

uses
  windows,sysutils,classes,ecma_type,ecma_expr;

//procedure InitProgram(var P: PJStatement);
//procedure FreeStatement(P: PJStatement);

type
  TJStatementFactory = class(TObject)
  private
    FList: TList;
    FExprFactory: TJExprFactory;
    procedure FreeStatement(P: PJStatement);
  public
    constructor Create(AExprFactory: TJExprFactory);
    destructor Destroy; override;
    procedure Clear;

    function NewStatement: PJStatement;
    function MakeExprStatement(Expr: PJExpr): PJStatement;
    //function MakePrintStatement(Expr: PJExpr): PJStatement;
    function MakeBlockStatement(Statements: PJStatement): PJStatement;
    function MakeReturnStatement(Expr: PJExpr): PJStatement;
    function MoveStart(P: PJStatement): PJStatement;
    function MoveLast(P: PJStatement): PJStatement;
    procedure MergeStatement(Prev,Next: PJStatement);
    function MakeIfStatement(Expr: PJExpr; TrueBlock: PJStatement; FalseBlock: PJStatement = nil): PJStatement;
    function MakeWhileStatement(Expr: PJExpr; Block: PJStatement): PJStatement;
    function MakeDoStatement(Expr: PJExpr; Block: PJStatement): PJStatement;
    function MakeForStatement(Expr1,Expr2,Expr3: PJExpr; Block: PJStatement): PJStatement;
    function MakeForInStatement(Expr1,Expr2: PJExpr; Block: PJStatement): PJStatement;
    function MakeContinueStatement: PJStatement;
    function MakeBreakStatement: PJStatement;
    function MakeWithStatement(Expr: PJExpr; Block: PJStatement): PJStatement;
    function MakeTryStatement(Block: PJStatement): PJStatement;
    function MakeCatchStatement(Expr: PJExpr; Block: PJStatement): PJStatement;
    function MakeFinallyStatement(Block: PJStatement): PJStatement;
    function MakeThrowStatement(Expr: PJExpr): PJStatement;
    function MakeEmptyStatement: PJStatement;
    function MakeVarDecl(Stmt: PJStatement): PJStatement;
    function MakeFunctionDecl(ID: PJExpr; ParamDecl,Block: PJStatement): PJStatement;
    function MakeParamDecl(Prev: PJStatement; ID: PJExpr): PJStatement;
    function MakeClassDecl(Name,Super: PJExpr; Block: PJStatement): PJStatement;
    function MakeVariableDecl(Name,Value: PJExpr): PJStatement;
    function MakeImportDecl(Name: PJExpr; SourceElements: PJStatement): PJStatement;
  end;

implementation

{
procedure InitProgram(var P: PJStatement);
//
begin
  if not Assigned(P) then
    Exit;

  FreeStatement(P);
  P := nil;
end;

procedure FreeStatement(P: PJStatement);
begin
  if not Assigned(P) then
    Exit;

  FreeStatement(P^.Next);
  FreeStatement(P^.Sub1);
  FreeStatement(P^.Sub2);

  InitTree(P^.Expr);
  Dispose(P);
end;
}

{ TJStatementFactory }

procedure TJStatementFactory.Clear;
//NA
var
  i: Integer;
begin
  for i := FList.Count - 1 downto 0 do
    FreeStatement(FList[i]);

  FList.Clear;
end;

constructor TJStatementFactory.Create(AExprFactory: TJExprFactory);
begin
  inherited Create;
  FList := TList.Create;
  FExprFactory := AExprFactory;
end;

destructor TJStatementFactory.Destroy;
begin
  Clear;
  FList.Free;
  inherited;
end;

procedure TJStatementFactory.FreeStatement(P: PJStatement);
begin
  Dispose(P);
end;  

function TJStatementFactory.NewStatement: PJStatement;
//VK쐬
begin
  New(Result);
  FillChar(Result^,SizeOf(TJStatement),0);
  Result^.SType := stNone;
  FList.Add(Result);
end;

function TJStatementFactory.MakeExprStatement(Expr: PJExpr): PJStatement;
//̍쐬
begin
  Result := NewStatement;
  Result^.Expr := Expr;
  Result^.SType := stExpr;
end;

{function TJStatementFactory.MakePrintStatement(Expr: PJExpr): PJStatement;
//print쐬
begin
  Result := NewStatement;
  Result^.Expr := Expr;
  Result^.SType := stPrint;
end;}

function TJStatementFactory.MakeBlockStatement(Statements: PJStatement): PJStatement;
//ubN쐬
// block - next(...)
// |-sub1(statements) - next(...)
begin
  Result := NewStatement;
  Result^.SType := stBlock;
  Result^.Sub1 := MoveStart(Statements);
end;

function TJStatementFactory.MakeParamDecl(Prev: PJStatement; ID: PJExpr): PJStatement;
//֐p[^錾
// param_decl(ID) - next(param_decl) - next(...)
//var
//  e: TJExpr;
begin
  Result := NewStatement;
  Result^.SType := stParamDecl;
  Result^.Expr := ID;
  //e := ID^;
  MergeStatement(Prev,Result);
end;

function TJStatementFactory.MakeFunctionDecl(ID: PJExpr; ParamDecl,Block: PJStatement): PJStatement;
//֐錾
// function_decl - next(...)
// |-sub1(paramdecl) - next(param_decl) - next(...)
// |-sub2(block)
//   |-sub1(statements) - next(...)
var
  ret: PJStatement;
begin
  Result := NewStatement;
  Result^.SType := stFunctionDecl;
  Result^.Sub1 := MoveStart(ParamDecl);
  Result^.Sub2 := MoveStart(Block);
  Result^.Expr := ID;
  //block̍Ōreturnt
  ret := MakeReturnStatement(FExprFactory.NewExpr);
  MergeStatement(MoveLast(Block),ret);
end;

function TJStatementFactory.MakeReturnStatement(Expr: PJExpr): PJStatement;
//return쐬
begin
  Result := NewStatement;
  Result^.SType := stReturn;
  Result^.Expr := Expr;
end;

function TJStatementFactory.MoveStart(P: PJStatement): PJStatement;
//ŏɈړ
begin
  Result := P;
  while Assigned(Result) do
  begin
    //eȂΏI
    if not Assigned(Result^.Parent) then
      Break
    else
      Result := Result^.Parent;
  end;
end;

function TJStatementFactory.MoveLast(P: PJStatement): PJStatement;
//ŌɈړ
begin
  Result := P;
  while Assigned(Result) do
  begin
    //ȂȂI
    if not Assigned(Result^.Next) then
      Break
    else
      Result := Result^.Next;
  end;
end;

procedure TJStatementFactory.MergeStatement(Prev,Next: PJStatement);
//݂ɂ  next <-> parent̂
begin
  if Assigned(Prev) then
    Prev^.Next := Next;

  if Assigned(Next) then
    Next^.Parent := Prev;
end;

function TJStatementFactory.MakeIfStatement(Expr: PJExpr; TrueBlock,FalseBlock: PJStatement): PJStatement;
//if쐬
// if   - next(...)
// |-sub1(trueblock)
// |-sub2(falseblock)
begin
  Result := NewStatement;
  Result^.SType := stIf;
  Result^.Expr := Expr;
  Result^.Sub1 := TrueBlock;
  Result^.Sub2 := FalseBlock;
end;

function TJStatementFactory.MakeWhileStatement(Expr: PJExpr; Block: PJStatement): PJStatement;
//while쐬
// while(expr)  - next(...)
// |-sub1(block)
begin
  Result := NewStatement;
  Result^.SType := stWhile;
  Result^.Expr := Expr;
  Result^.Sub1 := Block;
end;

function TJStatementFactory.MakeForStatement(Expr1,Expr2,Expr3: PJExpr; Block: PJStatement): PJStatement;
//for쐬
// for(expr1) --- next(...)
// |-sub1(block)
// |-sub2(expr2) --- next(expr3)
begin
  Result := NewStatement;
  Result^.SType := stFor;
  Result^.Expr := Expr1;
  Result^.Sub1 := Block;

  Result^.Sub2 := NewStatement;
  Result^.Sub2^.Expr := Expr2;

  Result^.Sub2^.Next := NewStatement;
  Result^.Sub2^.Next^.Expr := Expr3;
end;

function TJStatementFactory.MakeForInStatement(Expr1,Expr2: PJExpr; Block: PJStatement): PJStatement;
//for (in)쐬
// forin(expr1 - variable) --- next(...)
// |-sub1(block)
// |-sub2(expr2 - object)
begin
  Result := NewStatement;
  Result^.SType := stForIn;
  Result^.Expr := Expr1;
  Result^.Sub1 := Block;
  //IuWFNg
  Result^.Sub2 := NewStatement;
  Result^.Sub2^.Expr := Expr2;
end;

function TJStatementFactory.MakeContinueStatement: PJStatement;
//contnue쐬
begin
  Result := NewStatement;
  Result^.SType := stContinue;
end;

function TJStatementFactory.MakeBreakStatement: PJStatement;
//break쐬
begin
  Result := NewStatement;
  Result^.SType := stBreak;
end;

function TJStatementFactory.MakeWithStatement(Expr: PJExpr; Block: PJStatement): PJStatement;
//with쐬
// with(expr)  - next(...)
//  |
// sub1(block)
begin
  Result := NewStatement;
  Result^.SType := stWith;
  Result^.Expr := Expr;
  Result^.Sub1 := Block;
end;

function TJStatementFactory.MakeTryStatement(Block: PJStatement): PJStatement;
//try쐬
// try - next(...)
//  |
// sub1(block)
begin
  Result := NewStatement;
  Result^.SType := stTry;
  Result^.Sub1 := Block;
end;

function TJStatementFactory.MakeCatchStatement(Expr: PJExpr; Block: PJStatement): PJStatement;
//catch쐬
// catch(variable) - next(...)
//  |
// sub1(block)
begin
  Result := NewStatement;
  Result^.SType := stCatch;
  Result^.Expr := Expr;
  Result^.Sub1 := Block;
end;

function TJStatementFactory.MakeFinallyStatement(Block: PJStatement): PJStatement;
//finally쐬
// finally - next(...)
//  |
// sub1(block)
begin
  Result := NewStatement;
  Result^.SType := stFinally;
  Result^.Sub1 := Block;
end;

function TJStatementFactory.MakeThrowStatement(Expr: PJExpr): PJStatement;
//trow쐬
begin
  Result := NewStatement;
  Result^.SType := stThrow;
  Result^.Expr := Expr;
end;

function TJStatementFactory.MakeEmptyStatement: PJStatement;
//󕶂쐬
begin
  Result := NewStatement;
end;

function TJStatementFactory.MakeVarDecl(Stmt: PJStatement): PJStatement;
//var쐬
//var( ) - next(...)
// |- sub1(variable_declaration_list)
begin
  Result := NewStatement;
  Result^.SType := stVar;
  Result^.Sub1 := Stmt;
end;

function TJStatementFactory.MakeDoStatement(Expr: PJExpr;
  Block: PJStatement): PJStatement;
//do - while쐬
// do(expr)  - next(...)
// |-sub1(block)
begin
  Result := NewStatement;
  Result^.SType := stDo;
  Result^.Expr := Expr;
  Result^.Sub1 := Block;
end;

function TJStatementFactory.MakeClassDecl(Name, Super: PJExpr;
  Block: PJStatement): PJStatement;
//classdecl(expr: name,expr.left: super)  ... next(...)
// |- sub1(block)
begin
  Result := NewStatement;
  Result^.SType := stClassDecl;
  Result^.Expr := Name;
  Result^.Expr^.Left := Super;
  Result^.Sub1 := Block;
end;

function TJStatementFactory.MakeVariableDecl(Name,
  Value: PJExpr): PJStatement;
// memberdecl(expr: name,expr.left: value) - next(...)
begin
  Result := NewStatement;
  Result^.SType := stVarDecl;
  Result^.Expr := Name;
  Result^.Expr^.Left := Value;
end;

function TJStatementFactory.MakeImportDecl(Name: PJExpr; SourceElements: PJStatement): PJStatement;
// importdecl(expr: name)
//  | sub1(source_elements)
begin
  Result := NewStatement;
  Result^.SType := stImport;
  Result^.Expr := Name;
  Result^.Sub1 := SourceElements;
end;

end.
